#!/usr/bin/python
#
# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2007,2008, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.

"""
Separate the more general bump_analyze() from the other bump_*()
functions in calibcant.  Also provide a command line interface
for analyzing data acquired through other workflows.

The relevant physical quantities are :
 Vzp_out  Output z-piezo voltage (what we generate)
 Vzp      Applied z-piezo voltage (after external ZPGAIN)
 Zp       The z-piezo position
 Zcant    The cantilever vertical deflection
 Vphoto   The photodiode vertical deflection voltage (what we measure)

Which are related by the parameters :
 zpGain           Vzp_out / Vzp
 zpSensitivity    Zp / Vzp
 photoSensitivity Vphoto / Zcant

photoSensitivity is measured by bumping the cantilever against the
surface, where Zp = Zcant (see calibrate.bump_aquire()).  The measured
slope Vphoto/Vout is converted to photoSensitivity with bump_analyze().
"""

import numpy
import common # common module for the calibcant package
import config # config module for the calibcant package
import data_logger
import z_piezo_utils
import linfit
from splittable_kwargs import splittableKwargsFunction, \
    make_splittable_kwargs_function

#@splittableKwargsFunction((bump_plot, 'data'))
# Some of the child functions aren't yet defined, so postpone
# make-splittable until later in the module.
def bump_analyze(data, zpGain=config.zpGain,
                 zpSensitivity=config.zpSensitivity,
                 Vzp_out2V=config.Vzp_out2V,
                 Vphoto_in2V=config.Vphoto_in2V,
                 **kwargs) :
    """
    Return the slope of the bump ;).
    Inputs:
      data        dictionary of data in DAC/ADC bits
      Vzp_out2V   function that converts output DAC bits to Volts
      Vphoto_in2V function that converts input ADC bits to Volts
      zpGain      zpiezo applied voltage per output Volt
      zpSensitivity  nm zpiezo response per applied Volt
    Returns:
     photoSensitivity (Vphoto/Zcant) in Volts/nm
    Checks for strong correlation (r-value) and low randomness chance (p-value)
    
    With the current implementation, the data is regressed in DAC/ADC bits
    and THEN converted, so we're assuming that both conversions are LINEAR.
    if they aren't, rewrite to convert before the regression.
    """
    bump_plot_kwargs, = bump_analyze._splitargs(bump_analyze, kwargs)
    scale_Vzp_bits2V = Vzp_out2V(1) - Vzp_out2V(0)
    scale_Vphoto_bits2V = Vphoto_in2V(1) - Vphoto_in2V(0)
    Vphoto2Vzp_out_bit, intercept = \
        linfit.linregress(x=data["Z piezo output"],
                          y=data["Deflection input"],
                          plotVerbose=False)
    Vphoto2Vzp_out = Vphoto2Vzp_out_bit *scale_Vphoto_bits2V / scale_Vzp_bits2V

    #               1 / (Vzp/Vzp_out  *     Zp/Vzp       * Zcant/Zp )
    Vzp_out2Zcant = 1.0/ (zpGain      * zpSensitivity) # * 1
    bump_plot(data, **bump_plot_kwargs)
    return Vphoto2Vzp_out * Vzp_out2Zcant

@splittableKwargsFunction()
def bump_save(data, log_dir=None) :
    "Save the dictionary data, using data_logger.data_log()"
    if log_dir != None :
        log = data_logger.data_log(log_dir, noclobber_logsubdir=False,
                                   log_name="bump")
        log.write_dict_of_arrays(data)

def bump_load(datafile) :
    "Load the dictionary data, using data_logger.date_load()"
    dl = data_logger.data_load()
    data = dl.read_dict_of_arrays(datafile)
    return data

@splittableKwargsFunction()
def bump_plot(data, plotVerbose=False) :
    "Plot the bump (Vphoto vs Vzp) if plotVerbose or PYLAB_VERBOSE == True"
    if plotVerbose or config.PYLAB_VERBOSE :
        common._import_pylab()
        common._pylab.figure(config.BASE_FIGNUM)
        common._pylab.plot(data["Z piezo output"], data["Deflection input"],
                           '.', label='bump')
        common._pylab.title("bump surface")
        common._pylab.legend(loc='upper left')
        common._flush_plot()

make_splittable_kwargs_function(bump_analyze, (bump_plot, 'data'))

@splittableKwargsFunction((bump_analyze, 'data'))
def bump_load_analyze_tweaked(tweak_file, **kwargs):
    "Load the output file of tweak_calib_bump.sh, return an array of slopes"
    bump_analyze_kwargs, = \
        bump_load_analyze_tweaked._splitargs(bump_load_analyze_tweaked, kwargs)
    photoSensitivity = []
    for line in file(tweak_file, 'r') :
        parsed = line.split()
        path = parsed[0].strip()
        if path[0] == '#' : # a comment
            continue
        if config.TEXT_VERBOSE :
            print "Reading data from %s with ranges %s" % (path, parsed[1:])
        # read the data
        full_data = bump_load(path)
        if len(parsed) == 1 :
            data = full_data # use whole bump
        else :
            # use the listed sections
            zp = []
            df = []
            for rng in parsed[1:] :
                p = rng.split(':')
                starti = int(p[0])
                stopi = int(p[1])
                zp.extend(full_data['Z piezo output'][starti:stopi])
                df.extend(full_data['Deflection input'][starti:stopi])
            data = {'Z piezo output': numpy.array(zp),
                    'Deflection input': numpy.array(df)}
        pSi = bump_analyze(data, **bump_analyze_kwargs)
        photoSensitivity.append(pSi)
    return numpy.array(photoSensitivity, dtype=numpy.float)

# commandline interface functions
import scipy.io, sys

def read_data(ifile):
    "ifile can be a filename string or open (seekable) file object"
    if ifile == None :  ifile = sys.stdin
    unlabeled_data=scipy.io.read_array(ifile)
    data = {}
    data['Z piezo output'] = unlabeled_data[:,0]
    data['Deflection input'] = unlabeled_data[:,1]
    return data

def remove_further_than(data, zp_crit) :
    ndata = {}
    ndata['Z piezo output'] = []
    ndata['Deflection input'] = []
    for zp,df in zip(data['Z piezo output'],data['Deflection input']) :
        if zp > zp_crit :
            ndata['Z piezo output'].append(zp)
            ndata['Deflection input'].append(df)
    return ndata

if __name__ == '__main__' :
    # command line interface
    from optparse import OptionParser
    
    usage_string = ('%prog <input-file>\n'
                    '2008, W. Trevor King.\n'
                    '\n'
                    'There are two operation modes, one to analyze a single bump file,\n'
                    'and one to analyze tweak files.\n'
                    '\n'
                    'Single file mode (the default) :\n'
                    'Scales raw DAC/ADC bit data and fits a straight line.\n'
                    'Returns photodiode sensitivity Vphotodiode/Zcantilever in V/nm.\n'
                    '<input-file> should be whitespace-delimited, 2 column ASCII\n'
                    'without a header line.  e.g: "<zp_DAC>\\t<deflection_ADC>\\n"\n'
                    '\n'
                    'Tweak file mode:\n'
                    'Runs the same analysis as in single file mode for each bump in\n'
                    'a tweak file.  Each line in the tweak file specifies a single bump.\n'
                    'Blank lines and those beginning with a pound sign (#) are ignored.\n'
                    'The format of a line is a series of whitespace-separated fields--\n'
                    'a base file path followed by optional point index ranges, e.g.:\n'
                    '20080919/20080919132500_bump_surface 10:651 1413:2047\n'
                    'which only discards all points outside the index ranges [10,651)\n'
                    'and [1413,2047) (indexing starts at 0).\n'
                    )
    parser = OptionParser(usage=usage_string, version='%prog '+common.VERSION)
    parser.add_option('-C', '--cut-contact', dest='cut',
                      help='bilinear fit to cut out contact region (currently only available in single-file mode)',
                      action='store_true', default=False)
    parser.add_option('-o', '--output-file', dest='ofilename',
                      help='write output to FILE (default stdout)',
                      type='string', metavar='FILE')
    parser.add_option('-c', '--comma-out', dest='comma_out', action='store_true',
                      help='Output comma-seperated values (default %default)',
                      default=False)
    parser.add_option('-p', '--pylab', dest='pylab', action='store_true',
                      help='Produce pylab fit checks during execution',
                      default=False)
    parser.add_option('-t', '--tweak-mode', dest='tweakmode', action='store_true',
                      help='Run in tweak-file mode',
                      default=False)
    parser.add_option('-v', '--verbose', dest='verbose', action='store_true',
                      help='Print lots of debugging information',
                      default=False)

    options,args = parser.parse_args()
    parser.destroy()
    assert len(args) >= 1, "Need an input file"
        
    ifilename = args[0]

    if options.ofilename != None :
        ofile = file(options.ofilename, 'w')
    else :
        ofile = sys.stdout
    config.TEXT_VERBOSE = options.verbose
    config.PYLAB_INTERACTIVE = False
    config.PYLAB_VERBOSE = options.pylab
    config.GNUPLOT_VERBOSE = False

    if options.tweakmode == False :
        data = read_data(ifilename)
        if options.cut :
            ddict = {'approach':data} # although it may be any combination of approach and retract
            try :
                params = z_piezo_utils.analyzeSurfPosData(ddict, retAllParams=True)
                a,b,c,d = tuple(params) # model : f(x) = x<c ? a + b*x : (a+b*c) + d*(x-c)
                print >> sys.stderr, "fit with", params, ". using zp < %d" % c
                data = remove_further_than(data, c)
            except z_piezo_utils.poorFit, s :
                # data not very bilinear, so don't cut anything off.
                print >> sys.stderr, "use everything"
                
        photoSensitivity = bump_analyze(data)
        
        print >> ofile, photoSensitivity
    else : # tweak file mode
        slopes = bump_load_analyze_tweaked(ifilename)
        if options.comma_out :
            sep = ','
        else :
            sep = '\n'
        common.write_array(ofile, slopes, sep)
    
    if options.ofilename != None :
        ofile.close()
